#!/usr/bin/env python
# coding: utf-8

import json
import logging
import traceback
import importlib
import os
import time
import importlib.util
import sys
from intelliw.config.cfg_parser import load_config
from intelliw.datasets.datasets import DataSets
from intelliw.functions.feature_process import feature_process
from intelliw.utils.util import prepare_algorithm_parameters, prepare_model_parameters, import_code, get_first_element
from intelliw.utils.check_algorithm_py import is_valid_algorithm
import intelliw.utils.message as message
from intelliw.core.report import RestReport
from intelliw.core.recorder import Recorder
from intelliw.utils.logger import get_logger
from intelliw.config import config
from intelliw.utils.global_val import gl

logger = get_logger()


algorithm_py = "algorithm.py"
algorithm_yaml = "algorithm.yaml"
model_yaml = "model.yaml"
apihandler_py = "apihandler.py"

pre_predict = "pre-predict"
post_predict = "post-predict"
pre_train = "pre-train"
post_train = "post-train"

funcs_mapping = {
    "sys.general.getvalue": 'intelliw.functions.general.get_value',
    "sys.image.decode": 'intelliw.functions.opencv.im_decode',
    "sys.image.resize": 'intelliw.functions.opencv.im_resize',
    "sys.image.transpose": 'intelliw.functions.opencv.im_transpose',
    "sys.matrix.normalize": 'intelliw.functions.matrix.normalize',
    "sys.matrix.reshape": 'intelliw.functions.matrix.reshape',
    "sys.matrix.transpose": 'intelliw.functions.matrix.transpose',
    "sys.post.argmax": 'intelliw.functions.post.argmax',
    "sys.post.process": 'intelliw.functions.post.post_process'}


def get_model_yaml():
    return model_yaml


class PipelineException(Exception):
    pass


class Pipeline:
    def __init__(self, rep_addr=None, perodic_interval=-1):
        self.reporter = RestReport(rep_addr)
        self.transforms = None       # 针对总体数据结构处理的  特征工程
        self.perodic_interval = perodic_interval
        self.recorder = Recorder(self.reporter, self.perodic_interval <= 0)
        self.instance = None
        self.custom_router = []
        gl.set("reporter", self.reporter)

    def perodic_callback_train(self):
        logging.info("定时任务： 上报状态")
        self.recorder.report()

    def perodic_callback_infer(self):
        logging.info("定时任务： 上报状态")
        self.recorder.report_infer()

    def _check_algorithm_parameters_format(self, alg):
        try:
            alg_ps = alg['parameters'] if 'parameters' in alg else None
            if alg_ps is not None:
                for _, v in enumerate(alg_ps):
                    key = v['key'] if 'key' in v else None
                    option = v['option'] if 'option' in v else None
                    if option is None:
                        return False
            return True
        except:
            return False

    def _prepare(self, path):
        sys.path.append(path)
        py_file = os.path.join(path, algorithm_py)
        spec = importlib.util.spec_from_file_location("algorithm", py_file)
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)
        return self.module

    def _initialize(self, alg_file, model_file=None):
        try:
            cfg = load_config(alg_file)
            gl.set("alg_yaml_file", alg_file)

            # 路由配置获取
            self.custom_router = cfg['AlgorithmInformation'].get('router', [])

            # 算法配置获取
            alg_ps = cfg['AlgorithmInformation']['algorithm'].get(
                'parameters', None)
            parameters, option = prepare_algorithm_parameters(alg_ps)
            if model_file is not None:
                model_cfg = load_config(model_file)
                gl.set("model_yaml_file", model_file)
                self.model_cfg = model_cfg['Model']
                model_ps = self.model_cfg.get('parameters')
                # model and algorithm's parameters are all in this var 'parameters'
                parameters = prepare_model_parameters(
                    model_ps, parameters, option)

            # 特征处理可能会修改target_col，所以需要先保存起来
            target_col = parameters.get("targetCol")
            if target_col:
                gl.set("target_col", target_col)

            # 日志
            parameters["framework_log"] = logger

            instance = self.module.Algorithm(parameters)
            logger.info("algorithm instance {}".format(instance))

            if instance is None:
                self._raise_exception(message.err_import_alg_invalid_path)

            # 检测自定义路由方法是否存在
            for router in self.custom_router:
                func = router['func']
                if not hasattr(instance, func):
                    self._raise_exception(message.err_missing_router_func)

            # add decorator to report functions
            if config.is_server_mode():
                from intelliw.core.al_decorator import make_decorators
                make_decorators(instance, self.reporter)

            return instance
        except Exception as e:
            self._raise_exception(message.err_import_alg_invalid_path, e)

    def _importalg_phase1(self, path):
        if not os.path.isfile(os.path.join(path, algorithm_py)):
            self._raise_exception(message.err_import_alg_mssing_algorithpy)
        if not os.path.isfile(os.path.join(path, algorithm_yaml)):
            self._raise_exception(message.err_import_alg_mssing_algorithyaml)
        cfg = load_config(os.path.join(path, algorithm_yaml))
        self.algorithm_cfg = cfg                        # SELF-DEFINE DICT
        if not self._check_algorithm_parameters_format(cfg):
            self._raise_exception(
                message.err_import_alg_invalid_algorithyaml_format)
        if self._prepare(path) is None:
            self._raise_exception(
                message.err_import_alg_invalid_algorithpy_format)

        is_valid, msg = is_valid_algorithm(self.module)
        if not is_valid:
            self._raise_exception(msg)

    def importalg(self, path):
        """
        导入算法
        :param path: 算法包所在路径
        :return: None
        :raises PipelineException: 框架导入算法失败原因
        """
        try:
            self._importalg_phase1(path)
            self._initialize(os.path.join(path, algorithm_yaml))
        except Exception as e:
            self._raise_exception(message.err_import_algorithm, e)
        self.reporter.report(message.ok_import_algorithm)

    def importmodel(self, path, load_model=True):
        """
        导入模型
        :param path: 算法包所在路径
        :param load_model: 是否要导入模型，即是否调用 Algorithm 的 load 方法
        :return: None
        :raises PipelineException: 框架导入模型失败原因
        """
        if not os.path.isfile(os.path.join(path, model_yaml)):
            logger.error(
                "未找到 model.yaml, path: {}, file_name: {}".format(path, model_yaml))
            self._raise_exception(message.err_import_model_mssing_modelyaml)
        try:
            self._importalg_phase1(path)
        except Exception as e:
            self._raise_exception(message.err_import_model, e)
        self.instance = self._initialize(os.path.join(
            path, algorithm_yaml), os.path.join(path, model_yaml))
        try:
            trans_cfg = self.model_cfg.get('transforms', None)
            self.transforms, _feature_process = self.prepare_transforms(
                trans_cfg)
            gl.set('feature_process', _feature_process)
            if load_model:
                model_file = os.path.join(
                    path, self.model_cfg['location'])
                self.instance.load(model_file)
        except Exception as e:
            self._raise_exception(message.err_import_model, e)
        self.reporter.report(message.ok_import_model)

    def _transform_helper(self, stage, data):
        if self.transforms is not None:
            transforms = self.transforms.get(stage)
            if transforms:
                return self.transform_call_chains(transforms, data)
        return data

    def _transform_function(self, stage):
        def _transform(data):
            return self._transform_helper(stage, data)
        return _transform

    def _feature_helper(self,stage, data, _type=None):
        fp = gl.get('feature_process')
        if fp:
            params = fp.get(stage)
            if params:
             # 验证集数据应该按照infer来处理
                return feature_process(data, params, stage, _type)
        return data

    def _feature_process(self, stage):
        def _process(data, _type=None):
            return self._feature_helper(stage, data, _type)
        return _process

    def train(self, datasets: DataSets):
        try:
            if datasets is None:
                self._raise_exception(message.err_empty_train_data)

            # 训练集
            train_reader = datasets.train_reader(
                self._transform_function(pre_train), self._feature_process(pre_train))
            # 验证集
            valid_reader = datasets.validation_reader(
                self._transform_function(pre_train), self._feature_process(pre_train))

            # 开始训练
            train_func = getattr(self.instance, "train")
            if "kwargs" in train_func.__code__.co_varnames:
                train_func(train_reader, valid_reader,
                           final_target_col=gl.get("final_target_col"),
                           category_cols=gl.get("category_cols")
                           )
            else:
                train_func(train_reader, valid_reader)

            self.reporter.report(message.CommonResponse(
                200, 'train_finish', '', ''))
        except Exception as e:
            self._raise_exception(message.err_train, e)

    def infer(self, data, func='infer'):
        starttime = int(time.time() * 1000)
        try:
            logger.debug("receiving predict data {}".format(data))
            # 前处理
            data = self._transform_helper(pre_predict, data)
            data = self._feature_helper(pre_predict, data)

            # 推理
            infer_func = getattr(self.instance, func)
            if infer_func.__code__.co_argcount > 1:  # 获取函数参数数量
                data = infer_func(data)
            else:
                data = infer_func()

            # 后处理
            result = self._feature_helper(post_predict, data)
            result = self._transform_helper(post_predict, result)

            logger.debug("预测输出 {}".format(result))

            self.recorder.record_infer_status(
                '', True, starttime, int(time.time() * 1000))
            return result, None
        except Exception as e:
            stack_info = traceback.format_exc()
            logger.error("推理错误 {}, stack:\n{}".format(e, stack_info))
            self.recorder.record_infer_status('', False, starttime,
                                              int(time.time() * 1000), '{}'.format(e))
            return None, str("推理错误 {}".format(e))

    def transform_call_wrapper(self, src, func, params):
        if type(func) is type:
            return func().__call__(src, params)
        else:
            # treat as function type by default
            return func(src, params)

    def transform_call_chains(self, funcs, src):
        try:
            for _, item in enumerate(funcs):
                (name, func, params) = item
                src = self.transform_call_wrapper(src, func, params)
            return src
        except Exception as e:
            stack_info = traceback.format_exc()
            self.reporter.report(message.CommonResponse(500, "transforms",
                                                        "特征工程函数 {} 出错 {}, stack:\n{}".format(name, e, stack_info)))
            return src

    def _get_func_by_name(self, name):
        """
        根据 name 加载函数，支持格式: func_name or module_name.func_name

        加载顺序:
        1. 尝试在 self.module 中加载
        2. module_name.func_name 形式，尝试在指定的 module 中加载
        3. 尝试在当前 module 中加载

        :param name: 待加载的函数名称
        :return: 函数
        :raise: RuntimeError 加载失败原因
        """
        # 尝试在 self.module 中加载
        mod = self.module

        if hasattr(mod, name):
            return getattr(mod, name)

        # TODO 等ai工作坊修改后进行适配
        if "select_columns" in name:
            from intelliw.functions.select_columns import select_columns
            return select_columns
        if "filter_by_condition" in name:
            from intelliw.functions.data_filtering import filter_by_condition
            return filter_by_condition

        if name in funcs_mapping:
            name = funcs_mapping[name]
        
        # module_name.func_name 形式，尝试在指定的 module 中加载
        dot_pos = name.rfind('.')
        if dot_pos != -1:
            mod_name = name[0:dot_pos]
            func_name = name[dot_pos + 1:]
            if mod_name in sys.modules:
                mod = sys.modules[mod_name]
            else:
                try:
                    mod = importlib.import_module(mod_name)
                except ModuleNotFoundError:
                    raise RuntimeError(
                        '模块名无效: [{}], 函数无效: [{}]'.format(mod_name, name))
            if hasattr(mod, func_name):
                return getattr(mod, func_name)
            else:
                raise RuntimeError(
                    '函数名无效 [{}]，函数无效: [{}]'.format(func_name, name))

        # 尝试在当前 module 中加载
        mod = sys.modules[__name__]
        if hasattr(mod, name):
            return getattr(mod, name)

        raise RuntimeError('函数无效: [{}]'.format(name))

    def prepare_transforms(self, cfgs):
        transforms, features = dict(), dict()
        if cfgs:
            for v in cfgs:
                typ, functions = v.get('type'), v.get('functions')
                if typ is not None and functions is not None:
                    transforms[typ], features[typ] = list(), dict()
                    for item in functions:
                        key, parameters = item.get(
                            'key'), item.get('parameters')
                        params, _ = prepare_algorithm_parameters(parameters)
                        if key == 'feature_process':
                            features[typ] = json.loads(
                                params.get('feature_process_param', 'null'))
                        else:
                            body = item.get('body')
                            func = self._get_func_by_name(
                                key) if body is None else get_first_element(import_code(body, None))
                            if func is not None:
                                logger.info("特征函数 {}".format(key))
                                transforms[typ].append((key, func, params))
                            else:
                                logger.info("找不到函数 {}".format(key))
        return transforms, features

    def validate_transforms(self, cfgs, data):
        """
        验证特征工程函数， 用于用户校验函数正确性和结果，服务于validateService
        """
        transforms = []
        if cfgs is not None:
            for _, v in enumerate(cfgs):
                functions = v['functions'] if 'functions' in v else None
                if functions is not None:
                    for k, item in enumerate(functions):
                        desc = item['desc'] if 'desc' in item else None
                        key = item['key'] if 'key' in item else None
                        body = item['body'] if 'body' in item else None
                        parameters = item['parameters'] if 'parameters' in item else None
                        if body is not None:
                            nmod = import_code(body, None)
                            func = get_first_element(nmod)
                        else:
                            func = self._get_func_by_name(key)
                        params, _ = prepare_algorithm_parameters(parameters)
                        if func is not None:
                            logger.info("特征函数 {}".format(key))
                            transforms.append((key, func, params))
                        else:
                            logger.info("找不到函数 {}".format(key))

        return self.transform_call_chains(transforms, data)

    def _raise_exception(self, msg: message.CommonResponse, e: Exception = None):
        err_msg = "[{}]".format(msg.msg)
        # 用于传递错误给prepare_env.sh
        if e:
            err_stack = traceback.format_exc()
            err_msg += err_stack
        os.environ["ERR_MASSAGE"] = err_msg

        # e 为空，打印、上报、抛异常
        if e is None:
            logger.error(msg.msg)
            self.reporter.report(msg)
            raise PipelineException(msg.msg)

        # PipelineException 说明已经上报过了，打印过日志了，直接上抛
        if isinstance(e, PipelineException):
            raise e

        # 其他异常打印日志后上报错误
        logger.error(msg.msg + ' %s', e, exc_info=True)
        msg.msg = err_msg
        self.reporter.report(msg)

        raise PipelineException(msg.msg) from e
