import math
import random
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from intelliw.datasets import datasource_base

from intelliw.datasets.datasource_base import AbstractDataSource, DataSourceEmpty, AbstractDataSourceWriter, \
    EmptyDataSourceWriter
from intelliw.datasets.datasource_local_csv import DataSourceLocalCsv
from intelliw.utils.logger import get_logger
from intelliw.config import config

logger = get_logger()


class DataSourceType:
    '''输入数据源类型'''
    EMPTY = 0   # 空
    REMOTE_CSV = 1  # 远程csv
    INTELLIV = 2  # 智能分析
    LOCAL_CSV = 3  # 本地 csv
    IW_IMAGE_DATA = 4  # 图片数据源
    IW_FACTORY_DATA = 5  # 数据工场数据集


def get_datasource(intelliv_src: str, intelliv_row_addr: str) -> AbstractDataSource:
    datasource_type = config.SOURCE_TYPE
    if datasource_type == DataSourceType.EMPTY:
        return DataSourceEmpty()
    elif datasource_type == DataSourceType.REMOTE_CSV:
        from intelliw.datasets.datasource_remote_csv import DataSourceRemoteCsv
        return DataSourceRemoteCsv(config.DATA_SOURCE_ADDRESS)
    elif datasource_type == DataSourceType.INTELLIV:
        from intelliw.datasets.datasource_intelliv import DataSourceIntelliv
        return DataSourceIntelliv(intelliv_src, intelliv_row_addr, config.INPUT_MODEL_ID)
    elif datasource_type == DataSourceType.LOCAL_CSV:
        return DataSourceLocalCsv(config.CSV_PATH)
    elif datasource_type == DataSourceType.IW_IMAGE_DATA:
        from intelliw.datasets.datasource_iwimgdata import DataSourceIwImgData
        return DataSourceIwImgData(intelliv_src, intelliv_row_addr, config.INPUT_DATA_SOURCE_ID, config.INPUT_DATA_SOURCE_TRAIN_TYPE)
    elif datasource_type == DataSourceType.IW_FACTORY_DATA:
        from intelliw.datasets.datasource_iwfactorydata import DataSourceIwFactoryData
        return DataSourceIwFactoryData(intelliv_src, intelliv_row_addr, config.INPUT_DATA_SOURCE_ID)
    else:
        err_msg = "数据读取失败，无效的数据源类型: {}".format(datasource_type)
        logger.error(err_msg)
        raise ValueError(err_msg)


def get_datasource_writer(output_addr: str) -> AbstractDataSourceWriter:
    output_datasource_type = config.OUTPUT_SOURCE_TYPE
    if output_datasource_type == DataSourceType.EMPTY:
        return EmptyDataSourceWriter()
    elif output_datasource_type == DataSourceType.INTELLIV or output_datasource_type == DataSourceType.IW_FACTORY_DATA:
        from intelliw.datasets.datasource_intelliv import DataSourceWriter
        return DataSourceWriter(output_addr, config.OUTPUT_DATA_SOURCE_ID, config.INFER_ID, config.TENANT_ID)
    else:
        err_msg = "输出数据源设置失败，无效的数据源类型: {}".format(output_datasource_type)
        logger.error(err_msg)
        raise ValueError(err_msg)


class DataSets:
    def __init__(self, datasource: AbstractDataSource, config):
        self.datasource = datasource
        self.train_ratio = config.TRAIN_DATASET_RATIO

        if config.DATA_SPLIT_MODE == 0:
            spliter_cls = SequentialSpliter
        elif config.DATA_SPLIT_MODE == 1:
            spliter_cls = RandomSpliter
        else:
            err_msg = "输出数据源设置失败，数据集划分模式: {}".format(config.DATA_SPLIT_MODE)
            logger.error(err_msg)
            raise ValueError(err_msg)

        read_size = config.DATA_SOURCE_READ_SIZE
        if read_size == 0:
            read_size = 10000
        if read_size > config.DATA_SOURCE_READ_LIMIT > 0:
            read_size = config.DATA_SOURCE_READ_LIMIT

        self.spliter = spliter_cls(datasource, config.TRAIN_DATASET_RATIO, config.VALID_DATASET_RATIO,
                                   read_size, config.DATA_SOURCE_READ_LIMIT)

    def reader(self, page_size=100000, offset=0, limit=0, transform_function=None):
        return self.datasource.reader(page_size, offset, limit, transform_function)

    def train_reader(self, transform_function=None, feature_process=None):
        data = self.spliter.train_reader(transform_function)
        if feature_process:
            data = feature_process(data, "train_set")
        return data

    def validation_reader(self, transform_function=None, feature_process=None):
        data = self.spliter.validation_reader(transform_function)
        if feature_process:
            data = feature_process(data, "validation_set")
        return data

    @classmethod
    def read_all(cls, reader:Iterable) -> Iterable:
        # 图片数据为列表 其他数据集为{"result": , "meta": }
        alldata = list() if config.SOURCE_TYPE == DataSourceType.IW_IMAGE_DATA else dict()
        for data in reader:
            if config.SOURCE_TYPE == DataSourceType.IW_IMAGE_DATA:
                alldata.append(data)
            else:  
                if not alldata:
                    alldata = data
                elif 'result' in data and 'result' in alldata:
                    alldata['result'] = alldata['result'] + data['result']
        return alldata

def get_dataset(intelliv_src: str, intelliv_row_addr: str) -> DataSets:
    datasource = get_datasource(intelliv_src, intelliv_row_addr)
    return DataSets(datasource, config)


class DataSetSpliter(metaclass=ABCMeta):
    def __init__(self, datasource: AbstractDataSource, train_ratio, valid_ratio, read_size, read_limit):
        if not isinstance(datasource, AbstractDataSource):
            raise TypeError("data_source has a wrong type, required: AbstractDataSource, actually: {}"
                            .format(type(datasource).__name__))
        assert 0 < train_ratio <= 1
        assert 0 <= valid_ratio < 1
        assert read_size > 0

        self.datasource = datasource
        self.data_num = datasource.total()

        if read_limit == 0:
            self.total_read = self.data_num
        elif self.data_num > read_limit:
            self.total_read = read_limit
        else:
            self.total_read = self.data_num

        self.read_size = read_size
        if self.read_size > self.total_read:
            self.read_size = self.total_read

        self.train_num = math.floor(self.total_read * float(train_ratio))
        self.valid_num = self.total_read - self.train_num

    def train_reader(self, transform_function=None) -> Iterable:
        if self.train_num == 0:
            return DataSourceEmpty().reader()
        return self._train_reader(transform_function)

    @abstractmethod
    def _train_reader(self, transform_function=None) -> Iterable:
        pass

    def validation_reader(self, transform_function=None) -> Iterable:
        if self.valid_num == 0:
            return DataSourceEmpty().reader()
        return self._validation_reader(transform_function)

    @abstractmethod
    def _validation_reader(self, transform_function=None) -> Iterable:
        pass


class SequentialSpliter(DataSetSpliter):
    def __init__(self, datasource: AbstractDataSource, train_ratio, valid_ratio, read_size, read_limit):
        super().__init__(datasource, train_ratio, valid_ratio, read_size, read_limit)

    def _train_reader(self, transform_function=None):
        batch_size = self.read_size if self.read_size < self.train_num else self.train_num
        return self.datasource.reader(batch_size, 0, self.train_num, transform_function)

    def _validation_reader(self, transform_function=None):
        batch_size = self.read_size if self.read_size < self.valid_num else self.valid_num
        return self.datasource.reader(batch_size, self.train_num, self.valid_num, transform_function)


class RandomSpliter(DataSetSpliter):
    def __init__(self, datasource: AbstractDataSource, train_ratio, valid_ratio, read_size, read_limit):
        super().__init__(datasource, train_ratio, valid_ratio, read_size, read_limit)
        # 使用固定 seed 保证同一个数据集多次读取划分一致
        r = random.Random(1024)
        self.__mask = RandomSpliter.__build_mask(
            r, self.total_read, train_ratio)

    def _train_reader(self, transform_function=None) -> Iterable:
        train_musk_func = RandomSpliter.__mask_func(self.__mask, True)
        masked_transform = RandomSpliter.__masked_transform(
            train_musk_func, transform_function)
        return self.datasource.reader(self.read_size, 0, self.total_read, masked_transform)

    def _validation_reader(self, transform_function=None) -> Iterable:
        validation_musk_func = RandomSpliter.__mask_func(self.__mask, False)
        masked_transform = RandomSpliter.__masked_transform(
            validation_musk_func, transform_function)
        return self.datasource.reader(self.read_size, 0, self.total_read, masked_transform)

    @staticmethod
    def __masked_transform(mask_func, transform_function=None):
        def transform(result):
            result['result'] = mask_func(result['result'])
            if transform_function is not None:
                result = transform_function(result)
            return result

        return transform

    @staticmethod
    def __build_mask(r: random.Random, total_num: int, train_ratio: float):
        mask_size = total_num if total_num < 10000 else 1000
        mask = [False for _ in range(mask_size)]
        true_num = math.floor(len(mask) * train_ratio)
        for i in range(true_num):
            mask[i] = True
        r.shuffle(mask)
        return mask

    @staticmethod
    def __mask_func(mask: list, is_train: bool):
        index = [0]

        def func(data):
            res = []
            for val in data:
                if mask[index[0]] == is_train:
                    res.append(val)
                index[0] = index[0] + 1
                if index[0] >= len(mask):
                    index[0] = 0
            return res

        return func
