'''
Author: hexu
Date: 2022-02-20 10:49:18
LastEditTime: 2022-03-04 14:19:56
LastEditors: your name
Description: 按列的特征处理
FilePath: /iw-algo-fx/intelliw/functions/feature_process.py
'''
import json
from pyexpat import model
import numpy as np
import pandas as pd
from typing import Dict
from intelliw.utils import message
from intelliw.utils.util import default_dump
from collections import Iterable
from intelliw.datasets.datasets import DataSets
from intelliw.config import cfg_parser
from intelliw.config import config as intelliw_core_config
from intelliw.utils.logger import get_logger

from intelliw.utils.global_val import gl

logger = get_logger()

COl_PREFIX = "col_"

stage_feature_map = {
    "pre-train": "train",
    "post-train": "train",
    "pre-predict": "infer",
    "post-predict": "infer",
    "infer":"infer",
    "train":"train",
}

stage_infer_map = {
    "pre-train": "pre-predict",
    "post-train": "post-predict",
}

# injection_alg_yaml 新的配置注入到旧的配置里
# yaml_path 配置文件位置
# 能到这部key都是存在的，不用担心没有key
def injection_alg_yaml(yaml_path, val, stage, type):
    cfg = cfg_parser.load_single_config(yaml_path)
    value = [
                {
                    "key": "feature_process_param",
                    "desc": "特征处理参数",
                    "name": "特征处理参数",
                    "val": val,
                    "option": {"defaultValue": True, "type": "string"}
                }
            ]
    if type == "algorithm":
        transforms = cfg["AlgorithmInformation"]["algorithm"]["transforms"]
    else:
        transforms = cfg["Model"]["transforms"]
    for t in transforms:
        if t["type"] == stage:
            for f in t["functions"]:
                if f["key"] == "feature_process":
                    f["parameters"] = value
                    break
    else:
        transforms.append({"type":stage, "functions":[{"key":"feature_process", "parameters":value}]})
    return cfg

"""feature_process 特征工程处理
Args:
    data (_type_): 需要处理的数据
    config (Dict[str, dict]): 特征值配置
    stage (str): 处理类型 pre-train/post-train/pre-predict/post-predict
    _type (str): 数据集类型 验证集/训练集  如果胃验证集应按照推理的方式处理
Returns:
    Iterable: _description_
"""
def feature_process(data, config: Dict[str, dict], stage: str, _type:str) -> Iterable:
    alg_yaml_file, model_yaml_file = gl.get("alg_yaml_file"), gl.get("model_yaml_file")
    if not config and not alg_yaml_file and not model_yaml_file:
        return data

    logger.info("feature process start")
    processor_mode = stage_feature_map[stage]
    
    if processor_mode == "infer":
        df = pd.DataFrame(data["data"])
        fp = FeatureProcessor(config, processor_mode)
        result, _, _ = fp.do_process(df,fit=True if _type == "train_set" else False)
        data["data"] = result.values.tolist()
        return data
    else:
        # 获取所有数据
        alldata = DataSets.read_all(data)
        result = alldata.get("result")
        # 对数据进行处理
        if intelliw_core_config.SOURCE_TYPE != 4 and result:
            df = pd.DataFrame(result)
            # 特征处理可能会修改target_col，所以需要先保存起来
            target_col = gl.get("target_col")
            target_data = None
            if target_col and type(target_col) == int:
                target_data = df[df.columns[target_col]]
                logger.info("target_data save")

            fp = FeatureProcessor(config, processor_mode)
            result, category_cols, new_config = fp.do_process(df,True if _type == "train_set" else False)
            gl.set("category_cols", category_cols)
            
            # 将目标列加入数据
            if target_data is not None:
                result["new_target_data_unrepeatable_name"] = target_data
                gl.set("final_target_col", result.columns.get_loc("new_target_data_unrepeatable_name"))
                logger.info("final_target_col add")

            alldata["meta"] = [{'code': i} for i in result.columns.values.tolist()]
            alldata["result"] = result.values.tolist()
        
            logger.info("feature process end")
        else:
            return data
        
        # 如果是训练需要写入配置
        if _type == "train_set":
            gl.get('feature_process')[stage] = new_config
            new_config = json.dumps(new_config,ensure_ascii=False, default=default_dump)
            if not intelliw_core_config.is_server_mode():
                # 本地测试需要反写入配置文件
                model_yaml = injection_alg_yaml(model_yaml_file, new_config, stage_infer_map[stage], "model")
                cfg_parser.dump_config(model_yaml, model_yaml_file)
                alg_yaml = injection_alg_yaml(alg_yaml_file, new_config, stage_infer_map[stage], "algorithm")
                cfg_parser.dump_config(alg_yaml, alg_yaml_file)
            elif gl.get('reporter'):
                # 线上的上报到AI工作坊
                data = {"modelInstanceId": config.INSTANCE_ID,
                        "tenantId": config.TENANT_ID,
                        "valuationResult": {stage_infer_map[stage]:new_config}}
                gl.get('reporter').report(message.CommonResponse(200, 'feature_config', 'feature_config_upload', data))
            logger.info("injection yaml end")
        return [alldata]

class FeatureProcessor:
    def __init__(self, config_list, mode):
        self.config_list = config_list
        self.mode = mode       # train or infer
        self.l_Features = []
        if self.mode == 'train':
            for col_config in config_list:
                idx = col_config['col']
                if col_config['type'] in ['int', 'float', 'NUMBER']:
                    feat = NumericFeature(idx, col_config, self.mode)
                elif col_config['type'] in ['string', 'category', 'STRING', 'CATEGORY']:
                    feat = CategoryFeature(idx, col_config, self.mode)
                else:
                    assert False
                self.l_Features.append(feat)
        elif self.mode == 'infer':
            for idx, col_config in enumerate(config_list):
                if col_config['type'] in ['int', 'float', 'NUMBER']:
                    feat = NumericFeature(idx, col_config, self.mode)
                elif col_config['type'] in ['string', 'category', 'STRING', 'CATEGORY']:
                    feat = CategoryFeature(idx, col_config, self.mode)
                else:
                    assert False
                self.l_Features.append(feat)

    def do_process(self, data_frame, fit):
        '''
        :param data_frame:
        :return: when self.mode == 'train', processor will fit params with input data, if you just want to do data
                 process without change params, use set_mode('infer') before do_process!
        '''
        df = pd.DataFrame()
        self.category_cols = []

        if self.mode == 'train':
            if fit:
                self.__set_fit(True)
            else:
                self.__set_fit(False)
        else:
            self.__set_fit(False)

        for i, feat in enumerate(self.l_Features):
            col_num = df.shape[1]
            tmp_frame, res_params = feat.do_process(data_frame)
            if isinstance(feat, CategoryFeature):
                shape = tmp_frame.shape
                if len(shape) > 1:
                    self.category_cols.extend(range(col_num, col_num + shape[1]))
                else:
                    self.category_cols.append(col_num)
            df = pd.concat([df, tmp_frame], axis=1)
            if fit and self.mode == 'train':
                self.config_list[i]['func'] = res_params

        self.__set_fit(False)

        return df, self.category_cols, self.config_list


    def __set_fit(self, fit):
        if fit:
            for feat in self.l_Features:
                feat.mode = 'train'
        else:
            for feat in self.l_Features:
                feat.mode = 'infer'

class Feature:
    def __init__(self, index, config, mode, **kwargs):
        '''
        :param index:
        :param config:
        :param mode:
        :param auto_process: when mode == 'train' and auto_process == True the recommended func will be used, and params
                             will write back to the feature_engineer_config!
        :param auto_check_repair: when mode == 'train' and auto_check_repair == True, the recommended check_and_repair
                             method will be used, and params will write back to the feature_engineer_config!
        '''
        self.index = index
        self.mode = mode

        self.check_repair_config = config['check_repair']
        self.feature_engineer_config = config['func']
        self.fe_name2func = {}
        self.auto_check_repair = True if config.get('check_repair_auto', '0') == '1' else False
        self.auto_process = True if config.get('func_auto', '0') == '1' else False

    def do_process(self, data_frame):
        pass

class NumericFeature(Feature):
    def __init__(self, index, config, mode, **kwargs):
        super(NumericFeature, self).__init__(index, config, mode, **kwargs)
        self.fe_name2func = {'bucket': bucket,
                             'min_max_scale': min_max_scale,
                             'standard_scale': standard_scale,
                             'log_transform': log_transform,
                             'quantile_transform': quantile_transform
                             }


    def recommend_miss_repair(self, feat_df):
        return '1'

    def recommend_outlier_check(self, feat_df):
        return '1'

    def recommend_process(self, feat_df):
        return [{'name': 'min_max_scale', 'params': {}}]

    def do_process(self, data_frame):
        feat_df = data_frame[data_frame.columns[self.index]]

        # check and repair
        if self.mode == 'train':
            # missing repair
            assert 'miss_repair' in self.check_repair_config
            assert self.check_repair_config['miss_repair'] in ['1', '2', '3', '4', '5']
            feat_df.replace([np.inf, -np.inf, 'None', 'null'], np.nan, inplace=True)
            feat_df = feat_df.astype(float)
            if self.auto_check_repair:
                self.check_repair_config['miss_repair'] = self.recommend_miss_repair(feat_df)

            if self.check_repair_config['miss_repair'] == '1':
                feat_df = feat_df.fillna(feat_df.mean())
            elif self.check_repair_config['miss_repair'] == '2':
                feat_df = feat_df.fillna(0)
            elif self.check_repair_config['miss_repair'] == '3':
                feat_df = feat_df.fillna(feat_df.quantile(0.5))
            elif self.check_repair_config['miss_repair'] == '4':
                feat_df = feat_df.fillna(feat_df.mode()[0])

            if self.auto_check_repair:
                self.check_repair_config['outlier_check'] = self.recommend_outlier_check(feat_df)

            # outlier check and repair
            assert 'outlier_check' in self.check_repair_config
            # assert self.check_repair_config['outlier_check'] in ['1', '2', '3']
            assert self.check_repair_config['outlier_check'] == '1' or self.check_repair_config['outlier_check'] == '2' or self.check_repair_config['outlier_check'] == '3'
            # 3 西格玛
            if self.check_repair_config['outlier_check'] == '1':
                mean = feat_df.mean()
                std = feat_df.std()
                low = mean - 3 * std
                high = mean + 3 * std
                feat_df[feat_df < low] = low
                feat_df[feat_df > high] = high
            elif self.check_repair_config['outlier_check'] == '2':
                low = feat_df.quantile(0.25) - 1.5 * (feat_df.quantile(0.75) - feat_df.quantile(0.25))
                high = feat_df.quantile(0.75) + 1.5 * (feat_df.quantile(0.75) - feat_df.quantile(0.25))
                feat_df[feat_df < low] = low
                feat_df[feat_df > high] = high
        else:
            feat_df = feat_df.astype(float)

        if self.mode == 'train' and self.auto_process:
            self.feature_engineer_config = self.recommend_process(feat_df)

        # feature engineering
        for func in self.feature_engineer_config:
            feat_df, params = self.fe_name2func[func['name']](func['params'], feat_df, self.mode)
            if self.mode == 'train':
                func['params'] = params

        cols = []
        if len(feat_df.shape) == 1:
            cols = [f'{self.index}']

        elif len(feat_df.shape) == 2:
            for i in range(feat_df.shape[1]):
                cols.append(f'{self.index}_{i}')

        feat_df.columns = cols
        return feat_df, self.feature_engineer_config

class CategoryFeature(Feature):
    def __init__(self, index, config, mode, **kwargs):
        super(CategoryFeature, self).__init__(index, config, mode, **kwargs)

        self.fe_name2func = {'one_hot': one_hot,
                             'ordinal': ordinal
                             }

    def recommend_miss_repair(self, feat_df):
        return '4'

    def recommend_process(self, feat_df):
        return [{'name': 'one_hot', 'params': {}}]

    def do_process(self, data_frame):
        feat_df = data_frame[data_frame.columns[self.index]]
        # check and repair
        if self.mode == 'train':
            if self.auto_check_repair:
                self.feature_engineer_config = self.recommend_miss_repair(feat_df)
            # missing repair
            assert 'miss_repair' in self.check_repair_config
            assert self.check_repair_config['miss_repair'] in ['4', '5']
            if self.check_repair_config['miss_repair'] == '4':
                feat_df = feat_df.fillna(feat_df.mode()[0])

        if self.mode == 'train' and self.auto_process:
            self.feature_engineer_config = self.recommend_process(feat_df)

        # feature engineering
        for func in self.feature_engineer_config:
            # if func['name'] in self.fe_name2func:
            feat_df, params = self.fe_name2func[func['name']](func['params'], feat_df, self.mode)
            if self.mode == 'train':
                func['params'] = params

        cols = []
        if len(feat_df.shape) == 1:
            cols = [f'{self.index}']

        elif len(feat_df.shape) == 2:
            for i in range(feat_df.shape[1]):
                cols.append(f'{self.index}_{i}')

        feat_df.columns = cols
        return feat_df, self.feature_engineer_config

def ordinal(params, col_frame, mode):
    if mode == 'train':
        category_list = col_frame.value_counts().index.tolist()
        params['category_list'] = category_list
    else:
        category_list = params['category_list']
    col_frame[~col_frame.isin(category_list)] = len(category_list)
    for i, cate in enumerate(category_list):
        col_frame[col_frame == cate] = i
    col_frame = col_frame.astype(int)
    return col_frame, params

def one_hot(params, col_frame, mode):
    if mode == 'train':
        category_list = col_frame.value_counts().index.tolist()
        params['category_list'] = category_list
    else:
        category_list = params['category_list']

    col_one_hot = np.zeros((col_frame.shape[0], len(category_list)+1), dtype=np.int32 )

    for i, val in enumerate(col_frame.tolist()):
        if val == val and val in category_list: # not nan
            col_one_hot[i][category_list.index(val)] = 1
        else:
            col_one_hot[i][len(category_list)] = 1

    frame_new = pd.DataFrame(col_one_hot, index=col_frame.index)
    return frame_new, params

def quantile_transform(params, col_frame, mode):
    if mode == 'train':
        boundaries = []
        for pos in [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]:
            boundaries.append(col_frame[~col_frame.isnull()].quantile(pos))
        params['boundaries'] = boundaries
    else:
        boundaries = params['boundaries']

    bins = [float('-inf')]
    bins.extend(boundaries)
    bins.append(float('inf'))

    col_frame[~col_frame.isnull()] = pd.cut(col_frame[~col_frame.isnull()], bins, labels=False, duplicates='drop').astype(int)
    return col_frame, params

def log_transform(params, col_frame, mode):
    assert col_frame[(~col_frame.isnull()) & (col_frame < 0)].empty
    col_frame[~col_frame.isnull()] = np.log10(col_frame[~col_frame.isnull()]+1)
    return col_frame, params

def standard_scale(params, col_frame, mode):
    if mode == 'train':
        mean = col_frame[~col_frame.isnull()].mean()
        std = col_frame[~col_frame.isnull()].std()
        params['mean'] = mean
        params['std'] = std
    else:
        mean = params['mean']
        std = params['std']
    if std == 0:
        col_frame.iloc[:, ] = 1
    else:
        col_frame[~col_frame.isnull()] = (col_frame[~col_frame.isnull()] - mean) / std
    return col_frame, params

def max_abs_scale(params, col_frame, mode):
    if mode == 'train':
        max_abs = col_frame[~col_frame.isnull()].abs().max()
        params['max_abs'] = max_abs
    else:
        max_abs = params['max_abs']

    if max_abs == 0:
        col_frame.iloc[:, ] = 0
    else:
        col_frame[~col_frame.isnull()] = col_frame[~col_frame.isnull()]/max_abs
    return col_frame, params

def min_max_scale(params, col_frame, mode):
    if mode == 'train':
        max = col_frame[~col_frame.isnull()].max()
        min = col_frame[~col_frame.isnull()].min()
        params['max'] = max
        params['min'] = min
    else:
        max = params['max']
        min = params['min']
    if max == min:
        if max == 0:
            col_frame[~col_frame.isnull()].iloc[:, ] = 0
        else:
            col_frame[~col_frame.isnull()].iloc[:, ] = 1
    else:
        col_frame[~col_frame.isnull()] = (col_frame[~col_frame.isnull()] - min)/ (max - min)
    return col_frame, params

def bucket(params, col_frame, mode):
    assert 'boundaries' in params and len(params['boundaries']) > 0
    bins = [float('-inf')]
    bins.extend(params['boundaries'])
    bins.append(float('inf'))
    labels = range(len(bins)-1)
    col_frame[~col_frame.isnull()] = pd.cut(col_frame[~col_frame.isnull()], bins, labels=False, duplicates='drop').astype(int)
    return col_frame, params