"""
Intelliw debug tool suite
"""

import argparse
import os
import zipfile
import subprocess
import shutil
import platform
import intelliw.interface.controller as controller
from intelliw.utils.gen_model_cfg import generate_model_config_from_algorithm_config as __generate
from intelliw.config import config


def __parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("task", nargs="?", default="import_alg", type=str,
                        help="import_alg, import_model, train, infer, package, package_iwa, package_iwm; default is 'import_alg'")
    parser.add_argument("-p", "--path", default="packages",
                        type=str, help="project's path, default is 'packages'")
    parser.add_argument("--port", default=8888,
                        type=int, help="port to listen")
    parser.add_argument("-o", "--output_path", default="target",
                        type=str, help="package file output path")
    parser.add_argument("-f", '--csv', default="data.csv", type=str,
                        help="path of csv files as training data")
    parser.add_argument('--train_ratio', default=0.7, type=float,
                        help="train ratio, between 0.0 and 1.0, default is 0.7")
    parser.add_argument("-n", '--name', default="example",
                        type=str, help="algorithms name")
    return parser.parse_args()


def __args_to_framework_arg(__args, path):
    __framework_args = controller.FrameworkArgs()
    __framework_args.path = path

    if __args.task == 'import_alg':
        __framework_args.method = 'importalg'
    elif __args.task == 'import_model':
        __framework_args.method = 'importmodel'
    elif __args.task == 'train':
        __framework_args.method = 'batchservice'
        __framework_args.task = 'train'
    elif __args.task == 'infer':
        __framework_args.method = 'apiservice'
        __framework_args.port = __args.port
    else:
        raise ValueError(
            'failed to execute, unknown task: {}'.format(__args.task))

    return __framework_args


def __package(task: str, path: str, output_path: str):
    dir_name = os.path.split(path)[-1]
    if task == 'package' or task == 'package_iwa':
        file_name = dir_name + '.iwa'
    elif task == 'package_iwm':
        file_name = dir_name + '.iwm'
    elif task == 'package_iwp':
        file_name = dir_name + '.iwp'
    else:
        raise ValueError('Unknown package task: {}, supported task are package, package_iwa and package_iwm'
                         .format(task))
    output = os.path.join(output_path, file_name)

    if not os.path.exists(output_path):
        os.makedirs(output_path)

    if os.path.exists(output):
        os.remove(output)

    with zipfile.ZipFile(output, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dir, files in os.walk(path):
            if root.endswith('__pycache__'):
                continue
            relative_path = root.replace(path, "")
            if relative_path.startswith(os.sep):
                relative_path = relative_path[len(os.sep):]
            output_relative_path = os.path.join(dir_name, relative_path)
            for file in files:
                zipf.write(os.path.join(root, file),
                           os.path.join(output_relative_path, file))


def __init_algorithms(name: str, output_path: str):
    def alter(file, old_str, new_str):
        file_data = ""
        with open(file, "r", encoding="utf-8") as f:
            for line in f:
                if old_str in line:
                    line = line.replace(old_str, new_str)
                file_data += line
        with open(file, "w", encoding="utf-8") as f:
            f.write(file_data)

    if output_path == "target":
        output_path = os.getcwd()
    elif os.path.exists(output_path):
        os.chdir(output_path)
    else:
        raise FileNotFoundError(f"No such file or directory: '{output_path}'")

    if platform.system() == 'Windows':
        package_path = os.path.abspath(__file__).replace(
            "\\scaffolding\\resources\\iw.py", "")
    else:
        package_path = os.path.abspath(__file__).replace(
            "/scaffolding/resources/iw.py", "")
    algorithms_dir = os.path.join(package_path, "algorithms_demo")
    algorithm_yaml = os.path.join(name, "algorithm.yaml")
    pycache = os.path.join(name, "__pycache__")
    shutil.copytree(algorithms_dir, name)
    alter(algorithm_yaml, "AlgorithmName", name)
    if os.path.exists(pycache):
        shutil.rmtree(pycache)


def run():
    args = __parse_args()

    if args.task == "init":
        __init_algorithms(args.name, args.output_path)
        return

    _path = args.path
    if _path == 'packages':
        dirs = os.listdir(_path)
        if len(dirs) == 0:
            raise ValueError('empty packages')
        _path = os.path.join(_path, dirs[0])

    if args.task == 'package' or args.task == 'package_iwa' or args.task == 'package_iwm' or args.task == 'package_iwp':
        __package(args.task, _path, args.output_path)
    else:
        framework_args = __args_to_framework_arg(args, _path)

        if args.task == 'train':
            model_yaml_path = os.path.join(_path, 'model.yaml')
            algo_yaml_path = os.path.join(_path, 'algorithm.yaml')
            __generate(algo_yaml_path, model_yaml_path)

            assert 0.0 <= args.train_ratio <= 1.0
            config.TRAIN_DATASET_RATIO = args.train_ratio
            config.VALID_DATASET_RATIO = 1.0 - args.train_ratio
            config.DATA_SOURCE_READ_LIMIT = 20000000

            config.SOURCE_TYPE = 3
            config.CSV_PATH = args.csv

        config.RUNNING_MODE = 'SCAFFOLDING'
        controller.main(framework_args)


if __name__ == '__main__':
    run()
    # __init_algorithms()
