import os
import traceback
import requests
from intelliw.datasets.datasets import DataSets
from intelliw.interface.apijob import ApiService
from intelliw.core.infer import Infer
from intelliw.core.trainer import Trainer
from intelliw.utils.logger import get_logger
from intelliw.datasets.datasource_local_csv import DataSourceLocalCsv
from intelliw.utils.gen_model_cfg import generate_model_config_from_algorithm_config as __generate
from intelliw.config import config as intelliw_core_config_config

logger = get_logger()
report_addr = os.environ.get('REPORT_ADDR') or 'http://$local_ip:9001/data'
intelliw_core_config_config.RUNNING_MODE = 'SCAFFOLDING'
intelliw_core_config_config.SOURCE_TYPE = 3

#########################################################
#                请按照需求填写以下配置                     #
#########################################################

# (必须)设置算法文件位置（algorithms.py根目录）
PACKAGE_PATH = "/Users/hexer/Documents/workspace/code/iuapaipaas/iw-algo-fx/examples/dryrun"

algorithm_yaml_path = os.path.join(PACKAGE_PATH, 'algorithm.yaml')
model_yaml_path = os.path.join(PACKAGE_PATH, 'model.yaml')


# 训练测试(直接调用algorithm.train)
def test_train1():
    '''(test_train时必填)
    设置训练测试数据
    example: 
        TRAIN_CSV_PATH = "/home/xxx/data.csv"   训练数据
        EVAL_CSV_PATH = "/home/xxx/data.csv"    验证数据
    '''
    TRAIN_CSV_PATH = None
    EVAL_CSV_PATH = None

    try:
        __generate(algorithm_yaml_path,model_yaml_path)
        df_train = DataSourceLocalCsv(TRAIN_CSV_PATH).reader()
        df_eval = DataSourceLocalCsv(EVAL_CSV_PATH).reader()
        trainer = Trainer(PACKAGE_PATH)
        trainer.pipeline.instance.train(df_train, df_eval)
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error("fail to execute pipeline and stack:\n{}".format(str(stack_info)))


# 训练测试(通过算法框架调用train)
def test_train2():
    '''(test_train时必填)
    设置训练测试数据
    example: 
        TRAIN_CSV_PATH = "/home/xxx/data.csv"    训练数据
        intelliw_core_config_config.TRAIN_DATASET_RATIO = 0.7  训练数据比例
        intelliw_core_config_config.VALID_DATASET_RATIO = 0.3  验证数据比例
        intelliw_core_config_config.DATA_SPLIT_MODE = 0    数据集划分模式, 0 顺序划分,1 随机划分
    '''
    TRAIN_CSV_PATH = "/Users/hexer/Documents/workspace/code/iuapaipaas/iw-algo-fx/examples/dryrun/data.csv"
    intelliw_core_config_config.TRAIN_DATASET_RATIO = 0.7
    intelliw_core_config_config.VALID_DATASET_RATIO = 1 - intelliw_core_config_config.TRAIN_DATASET_RATIO
    intelliw_core_config_config.DATA_SPLIT_MODE = 0

    try:
        __generate(algorithm_yaml_path,model_yaml_path)

        train_data = DataSets(DataSourceLocalCsv(TRAIN_CSV_PATH), intelliw_core_config_config)
        trainer = Trainer(PACKAGE_PATH)
        trainer.train(train_data)
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error("fail to execute pipeline and stack:\n{}".format(str(stack_info)))


# 推理测试
def test_infer():
    '''(test_infer时必填)
    设置推理训练数据
    example: INFER_DATA = {"data":[1,2,3]}
    '''
    INFER_DATA = {"data":[[1.0 ,  "Alice" ,   77 ,   1.0 ,     1 ,   "A"   ,     1.00]]}
    {'data': [[0.0, 1.0, 77.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 9.900000990000098e-06]]}

    try:
        __generate(algorithm_yaml_path,model_yaml_path)
        infer = Infer(PACKAGE_PATH)
        result = infer.infer(INFER_DATA)
        print(result)
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error("fail to execute pipeline and stack:\n{}".format(str(stack_info)))



# 多api测试
def test_router():
    '''(test_router时必填)
    设置多api
    example:
        INFER_DATA = {"data":base64}
        ROUTER_FUNC = api_test
    '''
    INFER_DATA = None
    ROUTER_FUNC = None


    try:
        __generate(algorithm_yaml_path,model_yaml_path)
        infer = Infer(PACKAGE_PATH)
        instance = infer.pipeline.instance
        result = getattr(instance, ROUTER_FUNC)(INFER_DATA)
        print(result)
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error("fail to execute pipeline and stack:\n{}".format(str(stack_info)))


# http服务测试
def test_infer_http_server():
    # (test_infer_http_server时必填)设置http服务测试
    SERVER_PORT = 8088
    
    try:
        __generate(algorithm_yaml_path,model_yaml_path)
        apiservice = ApiService(SERVER_PORT, PACKAGE_PATH, report_addr)
        apiservice.run()
    except Exception as e:
        stack_info = traceback.format_exc()
        logger.error("fail to execute apiService and stack:\n{}".format(str(stack_info)))


# http接口测试
def test_server_available():
    '''(test_server_available时必填)
    http接口测试
    example：
        URL = "http://xxx.xxx.xxx/predict"
        METHOD = "get/post"
        INFER_DATA = {"data":base64}
    '''
    URL = "https://127.0.0.1:8888/predict"
    METHOD = "post"
    INFER_DATA = {'data': [[77068926790], [67068926790]]}

    resp = requests.request(METHOD, URL, json=INFER_DATA, timeout=60)
    if resp.status_code!=requests.codes.ok:
        resp.raise_for_status()
    print(resp.content)

if __name__ == '__main__':
    # test_train1()
    test_train2()
    # test_infer()
    # test_router()
    # test_infer_http_server()
    # test_server_available()
